import torch
from torch import nn
import numpy as np
from .layers import bn, VarianceLayer, CovarianceLayer, GrayscaleLayer
from .downsampler import * 
from torch.nn import functional
import torch.nn.functional as F
import math
from torch.autograd import Variable
from .L0_serial import *
import pytorch_ssim
import pytorch_msssim


class Gradient_L0Loss(nn.Module):
    def __init__(self):
        super(Gradient_L0Loss,self).__init__()
        self.diff = Diff()
        self.l1_loss = nn.L1Loss().type(torch.cuda.FloatTensor)

    def forward(self, R, mask):
        R_h = self.diff(R, 'h')
        R_w = self.diff(R, 'w')
       
        Rgrad = (R_h.abs() + R_w.abs())
        weight = (1-mask).pow(3)
        weight[weight<0.02] = 0
        # loss = weight * Rgrad.clamp(min=0.02).pow(-1)
        loss = weight * Rgrad
        return loss.mean()

class Gradient_L0LossV2(nn.Module):
    def __init__(self):
        super(Gradient_L0LossV2,self).__init__()
        self.diff = Diff()
        self.l1_loss = nn.L1Loss().type(torch.cuda.FloatTensor)

    def forward(self, S, mask):
        S_h = self.diff(S, 'h')
        S_w = self.diff(S, 'w')
        # L1:
        Sgrad = (S_h.abs() + S_w.abs())               
        # L2:
        # Sgrad = (S_h.abs() + S_w.abs()).pow(2)
        loss_S = mask * Sgrad
        # loss = weight * Rgrad.clamp(min=0.02).pow(-1)
        return loss_S.mean()

class Gradient_L0LossV3(nn.Module):
    def __init__(self):
        super(Gradient_L0LossV3,self).__init__()
        self.diff = Diff()
        self.l1_loss = nn.L1Loss().type(torch.cuda.FloatTensor)

    def forward(self, R, mask):
        R_h = self.diff(R, 'h')
        R_w = self.diff(R, 'w')
        # L1:
        Rgrad = (R_h.abs() + R_w.abs())/2               
        # L2:
        # Rgrad = (R_h.abs() + R_w.abs()).pow(2)                 
        loss_comple = mask * (1 - Rgrad).pow(2)        
        # loss = weight * Rgrad.clamp(min=0.02).pow(-1)
        return loss_comple.mean()


class WeightedGradientLoss(nn.Module):
    
    def __init__(self,threshold=0.02,alpha=1.2):
        super(WeightedGradientLoss,self).__init__()
        self.threshold = threshold
        self.alpha = alpha                      

    def generate_mask(self,x,thres):
        mask = torch.sign(x - thres)
        return mask.clamp(min=0)                 

    def forward(self,I,R,S):
        dif = Diff()
        I_h = dif(I,'h')
        I_w = dif(I,'w')
        R_h = dif(R,'h')
        R_w = dif(R,'w')
        S_h = dif(S,'h')
        S_w = dif(S,'w')

        epsilon = 1e-5
        I_log = torch.log(torch.clamp(I,min=epsilon,max=1.0))
        gradh_logI = dif(I_log,'h')
        gradw_logI = dif(I_log,'w') 
        weight_h = (gradh_logI.abs() ** self.alpha)          
        weight_w = (gradw_logI.abs() ** self.alpha)          

        maskI_h = self.generate_mask(I_h.mean(dim=1,keepdim=True).abs(), self.threshold)
        maskI_w = self.generate_mask(I_w.mean(dim=1,keepdim=True).abs(), self.threshold)
        maskR_h = self.generate_mask(R_h.mean(dim=1,keepdim=True).abs(), self.threshold)
        maskR_w = self.generate_mask(R_w.mean(dim=1,keepdim=True).abs(), self.threshold)
        maskS_h = self.generate_mask(S_h.mean(dim=1,keepdim=True).abs(), self.threshold)
        maskS_w = self.generate_mask(S_w.mean(dim=1,keepdim=True).abs(), self.threshold)
        
        mask_smooth_h = 1 - maskI_h
        mask_smooth_w = 1 - maskI_w
        
        loss_smooth = mask_smooth_h * (R_h.pow(2)+S_h.pow(2)) + mask_smooth_w * (S_w.pow(2)+R_w.pow(2)) 
     
        mask_h = maskR_h + maskS_h - maskI_h  
        mask_w = maskR_w + maskS_w - maskI_w                 
        loss_sw = mask_w.pow(2) * weight_w                   
        loss_sh = mask_h.pow(2) * weight_h

        return 0.5*(loss_sh + loss_sw).mean(), 0.5*loss_smooth.mean()       

class GradientLossV2(nn.Module):
    
    def __init__(self,threshold=0.02):
        super(GradientLossV2,self).__init__()
        self.threshold = threshold                     

    def generate_mask(self,x,thres):
        mask = torch.sign(x - thres)
        return mask.clamp(min=0)                 

    def forward(self,I,R,S):
        dif = Diff()
        I_h = dif(I,'h')
        I_w = dif(I,'w')
        R_h = dif(R,'h')
        R_w = dif(R,'w')
        S_h = dif(S,'h')
        S_w = dif(S,'w')

        maskI_h = self.generate_mask(I_h.mean(dim=1,keepdim=True).abs(), self.threshold)
        maskI_w = self.generate_mask(I_w.mean(dim=1,keepdim=True).abs(), self.threshold)
        
        mask_smooth_h = 1 - maskI_h
        mask_smooth_w = 1 - maskI_w
       
        loss_smooth = mask_smooth_h * (R_h.pow(2)+S_h.pow(2)) + mask_smooth_w * (S_w.pow(2)+R_w.pow(2)) 
        #loss_smooth = mask_smooth_h *(maskR_h + maskS_h)+ mask_smooth_w * (maskR_w + maskS_w)
        
        exclu = ExclusionLoss()
        loss_exclu = exclu(R,S)
        return loss_exclu, 0.5*loss_smooth.mean() 


class WeightedGradientConsistLoss(nn.Module):
    def __init__(self,threshold=0.02,alpha=1.2):
        super(WeightedGradientConsistLoss,self).__init__()
        self.threshold = threshold                             

    def generate_mask(self,x,thres):
        mask = torch.sign(x - thres)
        return mask.clamp(min=0)                   

    def forward(self,I,R,S):
        dif = Diff()
        I_h = dif(I,'h')
        I_w = dif(I,'w')
        R_h = dif(R,'h')
        R_w = dif(R,'w')        
        S_h = dif(S,'h')
        S_w = dif(S,'w')

        epsilon = 1e-4
        I_log = torch.log(torch.clamp(I,min=epsilon,max=1.0))
        gradh_logI = dif(I_log,'h')
        gradw_logI = dif(I_log,'w')
        weight_h = (gradh_logI.abs() ** self.alpha + epsilon)**(-1)
        weight_w = (gradw_logI.abs() ** self.alpha + epsilon)**(-1)
        
        maskI_h = self.generate_mask(I_h.abs(), self.threshold)
        maskI_w = self.generate_mask(I_w.abs(), self.threshold)
        maskRS_h = self.generate_mask(R_h.abs() + S_h.abs(), self.threshold )  
        maskRS_w = self.generate_mask(R_w.abs() + S_w.abs(), self.threshold )  
        
        mask_h = maskRS_h - maskI_h  
        mask_w = maskRS_w - maskI_w                   

        loss_dh = mask_h.clamp(min=0) * (R_h.pow(2)+ S_h.pow(2))
        loss_sh = (-mask_h).clamp(min=0) * I_h.pow(2) 
        loss_dw = mask_w.clamp(min=0) * (R_w.pow(2) + S_w.pow(2))
        loss_sw = (-mask_w).clamp(min=0) * I_w.pow(2)
        # smooth loss
        mask_smooth_h = 1 - maskI_h
        mask_smooth_w = 1 - maskI_w
        loss_smooth = mask_smooth_h * (R_h.pow(2)+S_h.pow(2)) + mask_smooth_w * (S_w.pow(2)+R_w.pow(2))        
        return 0.5*(weight_h*(loss_dh+loss_sh) + weight_w*(loss_dw + loss_sw)).mean(), 0.5*loss_smooth.mean()

class GradientConsistLoss(nn.Module):
    def __init__(self,threshold=0.02):
        super(GradientConsistLoss,self).__init__()
        self.threshold = threshold                             

    def generate_mask(self,x,thres):
        mask = torch.sign(x - thres)
        return mask.clamp(min=0)                   

    def forward(self,I,R,S):
        dif = Diff()
        I_h = dif(I,'h')
        I_w = dif(I,'w')
        R_h = dif(R,'h')
        R_w = dif(R,'w')        
        S_h = dif(S,'h')
        S_w = dif(S,'w')
        
        maskI_h = self.generate_mask(I_h.abs(), self.threshold)
        maskI_w = self.generate_mask(I_w.abs(), self.threshold)
        maskRS_h = self.generate_mask(R_h.abs() + S_h.abs(), self.threshold ) 
        maskRS_w = self.generate_mask(R_w.abs() + S_w.abs(), self.threshold )  
       
        mask_h = maskRS_h - maskI_h  
        mask_w = maskRS_w - maskI_w                   

        loss_dh = mask_h.clamp(min=0) * (R_h.pow(2)+ S_h.pow(2))
        loss_sh = (-mask_h).clamp(min=0) * I_h.pow(2) 
        loss_dw = mask_w.clamp(min=0) * (R_w.pow(2) + S_w.pow(2))
        loss_sw = (-mask_w).clamp(min=0) * I_w.pow(2)
        # smooth loss
        mask_smooth_h = 1 - maskI_h
        mask_smooth_w = 1 - maskI_w
        loss_smooth = mask_smooth_h * (R_h.pow(2)+S_h.pow(2)) + mask_smooth_w * (S_w.pow(2)+R_w.pow(2))        
        return 0.5*(loss_dh + loss_dw + loss_sh + loss_sw).mean(), 0.5*loss_smooth.mean()


class HistogramLoss(nn.Module):
    def __init__(self):
        super(HistogramLoss,self).__init__()
    def forward(self,I,S):
        I_gray = I[:,0,:,:] * 0.299 + I[:,1,:,:] * 0.587 + I[:,2,:,:] * 0.114
        I_list = []
        S_list = []
        temp_I = 0
        temp_S = 0
        for i in range(1,10):
            thres = 1 - 0.1 * i
            p_I = torch.sign(I_gray - thres).clamp(min=0).mean() - temp_I
            p_S = torch.sign(S - thres).clamp(min=0).mean() - temp_S
            I_list.insert(0,p_I)                                    
            S_list.insert(0,p_S)
            temp_I = temp_I + p_I
            temp_S = temp_S + p_S

        I_list.insert(0, 1-sum(I_list))
        S_list.insert(0, 1-sum(S_list))                            
        p_target = torch.stack(I_list,0)                         
        p_estimate = torch.stack(S_list,0)
        ##     https://github.com/truskovskiyk/nima.pytorch/blob/master/nima/emd_loss.py
        assert p_target.shape == p_estimate.shape
        # cdf for values [1, 2, ..., 10]
        cdf_target = torch.cumsum(p_target, dim=0)             
        # cdf for values [1, 2, ..., 10]
        cdf_estimate = torch.cumsum(p_estimate, dim=0)
        cdf_diff = cdf_estimate - cdf_target
        samplewise_emd = torch.sqrt(torch.mean(torch.pow(torch.abs(cdf_diff), 2)))
        return samplewise_emd.mean()

class ColorLoss(nn.Module):
    def __init__(self):
        super(ColorLoss,self).__init__()
    def forward(self,I,R):
        eps = 1e-6
        
        R = R / R.sum(dim=1,keepdim=True).clamp(min=eps)
        I = I / I.sum(dim=1,keepdim=True).clamp(min=eps)
        denorm1 = R.pow(2).sum(dim=1,keepdim=True).clamp(min=eps)      
        denorm2 = I.pow(2).sum(dim=1,keepdim=True).clamp(min=eps)  
        denorm = (denorm1 * denorm2).pow(0.5) + eps 
                                                          
        cos_temp = (I*R).sum(dim=1,keepdim=True) / denorm
        cos = cos_temp.clamp(min=-0.999999, max=0.999999)
      
        angel = torch.acos(cos).mean()           
        
        # cos.register_hook(lambda g: print('\ncos grad -----------------',torch.isnan(g).max()))

        # denorm1_1.register_hook(lambda g: print('\ndenorm1_1 grad -----------------',torch.isnan(g).max()))
        # denorm1_2.register_hook(lambda g: print('\ndenorm1_2 grad -----------------',torch.isnan(g).max()))
        # denorm1.register_hook(lambda g: print('\ndenorm1 grad -----------------',torch.isnan(g).max()))

        # # denorm2.register_hook(lambda g: print('\ndenorm2 grad -----------------',torch.isnan(g).max()))
        # denorm.register_hook(lambda g: print('\ndenorm grad -----------------',torch.isnan(g).max()))
        # R.register_hook(lambda g: print('\nR grad -----------------',torch.isnan(g).max()))
        # R_temp.register_hook(lambda g: print('\nR_temp grad -----------------',torch.isnan(g).max()))
        # R_temp2.register_hook(lambda g: print('\nR_temp2 grad -----------------',torch.isnan(g).max()))
        # angel.register_hook(lambda g: print(g))
        return angel


class StdLoss(nn.Module):
    def __init__(self):
        """
        Loss on the variance of the image.
        Works in the grayscale.
        If the image is smooth, gets zero
        """
        super(StdLoss, self).__init__()
        blur = (1 / 25) * np.ones((5, 5))
        blur = blur.reshape(1, 1, blur.shape[0], blur.shape[1])
        self.mse = nn.MSELoss()
        self.blur = nn.Parameter(data=torch.cuda.FloatTensor(blur), requires_grad=False)
        image = np.zeros((5, 5))
        image[2, 2] = 1
        image = image.reshape(1, 1, image.shape[0], image.shape[1])
        self.image = nn.Parameter(data=torch.cuda.FloatTensor(image), requires_grad=False)
        self.gray_scale = GrayscaleLayer()

    def forward(self, x):
        x = self.gray_scale(x)
        return self.mse(functional.conv2d(x, self.image), functional.conv2d(x, self.blur))

class TV2Loss(nn.Module):
    def __init__(self):
        super(TV2Loss, self).__init__()
    def forward(self,x):
        diff = Diff()
        grad_h = diff(x, 'h')
        grad_w = diff(x, 'w')
        return (grad_h.pow(2) + grad_w.pow(2)).mean()

class TV1Loss(nn.Module):
    def __init__(self):
        super(TV1Loss, self).__init__()
    def forward(self,x):
        diff = Diff()
        grad_h = diff(x, 'h')
        grad_w = diff(x, 'w')
        return (grad_h.abs() + grad_w.abs()).mean()

class IntensityLoss(nn.Module):
    def __init__(self):
        super(IntensityLoss,self).__init__()

    def forward(self,x,rate):                             
        # y = torch.clamp(x-0.8,min=0)
        # return (torch.exp(y*weight)).mean()
        return (torch.exp(x*rate)).mean()


class ExclusionLoss(nn.Module):

    def __init__(self, iflog=False, log_eps=0.001,level=3):
        
        super(ExclusionLoss, self).__init__()
        self.level = level
        self.iflog = iflog
        self.log_eps = log_eps
        self.avg_pool = torch.nn.AvgPool2d(2, stride=2).type(torch.cuda.FloatTensor)
        self.sigmoid = nn.Sigmoid().type(torch.cuda.FloatTensor)

    def get_gradients(self, img1, img2):
        gradx_loss = []
        grady_loss = []

        for l in range(self.level):
            if self.iflog:
                gradx1, grady1 = self.compute_gradient(torch.log(img1.clamp(min=self.log_eps)))
                gradx2, grady2 = self.compute_gradient(torch.log(img2.clamp(min=self.log_eps)))
            else:
                gradx1, grady1 = self.compute_gradient(img1)
                gradx2, grady2 = self.compute_gradient(img2)

            alphay = 1
            alphax = 1
            ##################################################################################
            gradx1_s = (self.sigmoid(gradx1) * 2) - 1                
            grady1_s = (self.sigmoid(grady1) * 2) - 1
            gradx2_s = (self.sigmoid(gradx2 * alphax) * 2) - 1
            grady2_s = (self.sigmoid(grady2 * alphay) * 2) - 1

            gradx_loss += self._all_comb(gradx1_s, gradx2_s)
            grady_loss += self._all_comb(grady1_s, grady2_s)
            img1 = self.avg_pool(img1)
            img2 = self.avg_pool(img2)
        return gradx_loss, grady_loss
    

   
    def _all_comb(self, grad1_s, grad2_s):
        v = []
        for i in range(grad2_s.size()[1]):
            for j in range(grad1_s.size()[1]):
                v.append(torch.mean(((grad1_s[:, j, :, :] ** 2) * (grad2_s[:, i, :, :] ** 2))) ** 0.25)
        return v

    def forward(self, img1, img2):
        
        k = img1.size()[1] * img2.size()[1]                             
        #k = 9
        gradx_loss, grady_loss = self.get_gradients(img1, img2)
        loss_gradxy = sum(gradx_loss) / (self.level * 9) + sum(grady_loss) / (self.level * 9)
        return loss_gradxy / 2.0

    def compute_gradient(self, img):
        gradx = img[:, :, 1:, :] - img[:, :, :-1, :]
        # gradx = F.pad(gradx,(0,0,0,1))
        grady = img[:, :, :, 1:] - img[:, :, :, :-1]
        # grady = F.pad(grady,(0,1,0,0))
        return gradx, grady


class ExtendedL1Loss(nn.Module):
    """
    also pays attention to the mask, to be relative to its size
    """
    def __init__(self):
        super(ExtendedL1Loss, self).__init__()
        self.l1 = nn.L1Loss().cuda()

    def forward(self, a, b, mask):
        normalizer = self.l1(mask, torch.zeros(mask.shape).cuda())
        # if normalizer < 0.1:
        #     normalizer = 0.1
        c = self.l1(mask * a, mask * b) / normalizer
        return c


class NonBlurryLoss(nn.Module):
    def __init__(self):
        
        super(NonBlurryLoss, self).__init__()
        self.mse = nn.MSELoss()

    def forward(self, x):
        return 1 - self.mse(x, torch.ones_like(x) * 0.5)


class GrayscaleLoss(nn.Module):
    def __init__(self):
        super(GrayscaleLoss, self).__init__()
        self.gray_scale = GrayscaleLayer()
        self.mse = nn.MSELoss().cuda()

    def forward(self, x, y):
        x_g = self.gray_scale(x)
        y_g = self.gray_scale(y)
        return self.mse(x_g, y_g)


class GrayLoss(nn.Module):
    def __init__(self):
        super(GrayLoss, self).__init__()
        self.l1 = nn.L1Loss().cuda()

    def forward(self, x):
        y = torch.ones_like(x) / 2.
        return 1 / self.l1(x, y)


# class GradientLoss(nn.Module):
#     """
#     L1 loss on the gradient of the picture
#     """
#     def __init__(self):
#         super(GradientLoss, self).__init__()

#     def forward(self, a):
#         gradient_a_x = torch.abs(a[:, :, :, :-1] - a[:, :, :, 1:])
#         gradient_a_y = torch.abs(a[:, :, :-1, :] - a[:, :, 1:, :])
#         return torch.mean(gradient_a_x) + torch.mean(gradient_a_y)


class YIQGNGCLoss(nn.Module):
    def __init__(self, shape=5):
        super(YIQGNGCLoss, self).__init__()
        self.shape = shape
        self.var = VarianceLayer(self.shape, channels=1)
        self.covar = CovarianceLayer(self.shape, channels=1)

    def forward(self, x, y):
        if x.shape[1] == 3:
            x_g = rgb_to_yiq(x)[:, :1, :, :]  # take the Y part
            y_g = rgb_to_yiq(y)[:, :1, :, :]  # take the Y part
        else:
            assert x.shape[1] == 1
            x_g = x  # take the Y part
            y_g = y  # take the Y part
        c = torch.mean(self.covar(x_g, y_g) ** 2)
        vv = torch.mean(self.var(x_g) * self.var(y_g))
        return c / vv
####################################################################################################################
#############################################     RTV Loss    ######################################################
class GaussianFilter(nn.Module):
    def __init__(self):
        super(GaussianFilter,self).__init__()
        kernel_list_1 = [0.0545,0.2442,0.4026,0.2442,0.0545]
        self.kernel_1 = torch.FloatTensor(kernel_list_1).unsqueeze(0).unsqueeze(0).unsqueeze(0) 
        self.kernel_tran_1 = torch.FloatTensor(kernel_list_1).unsqueeze(0).unsqueeze(0).unsqueeze(3)

        kernel_list_15 = [0.0366,0.1113,0.2167,0.2707,0.2167,0.1113,0.0366]
        self.kernel_15 = torch.FloatTensor(kernel_list_15).unsqueeze(0).unsqueeze(0).unsqueeze(0) 
        self.kernel_tran_15 = torch.FloatTensor(kernel_list_15).unsqueeze(0).unsqueeze(0).unsqueeze(3)

        # kernel_list_2 = [0.0088,0.0271,0.0651,0.1216,0.1770,0.2006,0.1770,0.1216,0.0651,0.0271,0.0088]
        # 下面这个 list 是特别为 MultiscaleDIP inference 阶段的推测设计的
        kernel_list_2 = [0.0276,0.0663,0.1238,0.1802,0.2042,0.1802,0.1238,0.0663,0.0276]
        self.kernel_2 = torch.FloatTensor(kernel_list_2).unsqueeze(0).unsqueeze(0).unsqueeze(0) 
        self.kernel_tran_2 = torch.FloatTensor(kernel_list_2).unsqueeze(0).unsqueeze(0).unsqueeze(3)

        kernel_list_25 = [0.0185,0.0342,0.0563,0.0831,0.1097,0.1296,0.1370,0.1296,0.1097,0.0831,0.0563,0.0342,0.0185]
        self.kernel_25 = torch.FloatTensor(kernel_list_25).unsqueeze(0).unsqueeze(0).unsqueeze(0)
        self.kernel_tran_25 = torch.FloatTensor(kernel_list_25).unsqueeze(0).unsqueeze(0).unsqueeze(3)

        # RTV sigma = 3.0  padding = (15 - 1)/2 = 7
        kernel_list_3 = [0.0088, 0.0182, 0.0336, 0.0553, 0.0816, 0.1078, 0.1273, 0.1346, 0.1273, 0.1078, 0.0816, 0.0553, 0.0336, 0.0182, 0.0088]
        self.kernel_3 = torch.FloatTensor(kernel_list_3).unsqueeze(0).unsqueeze(0).unsqueeze(0) 
        self.kernel_tran_3 = torch.FloatTensor(kernel_list_3).unsqueeze(0).unsqueeze(0).unsqueeze(3)
        # self.register_buffer('kernel_3', kernel_3)
        # self.register_buffer('kernel_tran_3', kernel_tran_3)

        # RTV sigma = 4.0  padding = (21 - 1)/2 = 10
        kernel_list_4 = [0.0044, 0.0080, 0.0136, 0.0218, 0.0327, 0.0461, 0.0610, 0.0759, 0.0888, 0.0975, 0.1006, \
            0.0975, 0.0888, 0.0759,0.0610, 0.0461, 0.0327,0.0218,0.0136,0.0080,0.0044]
        self.kernel_4 = torch.FloatTensor(kernel_list_4).unsqueeze(0).unsqueeze(0).unsqueeze(0) 
        self.kernel_tran_4 = torch.FloatTensor(kernel_list_4).unsqueeze(0).unsqueeze(0).unsqueeze(3)

        # RTV sigma = 5.0  padding = (25 - 1)/2 = 12
        kernel_list_5 = [0.0045,0.0072,0.0109,0.0160,0.0225,0.0303,0.0393,0.0490,0.0587,0.0675 ,0.0746,0.0792,0.0808,\
            0.0792,0.0746,0.0675,0.0587,0.0490,0.0393,0.0303,0.0225,0.0160,0.0109,0.0072,0.0045]
        self.kernel_5 = torch.FloatTensor(kernel_list_5).unsqueeze(0).unsqueeze(0).unsqueeze(0) 
        self.kernel_tran_5 = torch.FloatTensor(kernel_list_5).unsqueeze(0).unsqueeze(0).unsqueeze(3)

        # RTV sigma = 6.0  padding = (31 - 1)/2 = 15
        kernel_list_6 = [0.0029,0.0044,0.0064,0.0091,0.0125,0.0167,0.0218,0.0276,0.0340,0.0407,0.0474,0.0538,0.0593,0.0635,0.0662,0.0671,0.0662,\
            0.0635,0.0593,0.0538,0.0474,0.0407,0.0340,0.0276 ,0.0218,0.0167,0.0125 ,0.0091 ,0.0064,0.0044,0.0029]
        self.kernel_6 = torch.FloatTensor(kernel_list_6).unsqueeze(0).unsqueeze(0).unsqueeze(0) 
        self.kernel_tran_6 = torch.FloatTensor(kernel_list_6).unsqueeze(0).unsqueeze(0).unsqueeze(3)



    def forward(self, x, sigmma):
        if sigmma == 2.5:
            self.kernel = self.kernel_25.to(x.device)
            self.kernel_tran = self.kernel_tran_25.to(x.device)
            self.n_padding = 6
        elif sigmma == 3.0:
            self.kernel = self.kernel_3.to(x.device)
            self.kernel_tran = self.kernel_tran_3.to(x.device)
            self.n_padding = 7
        elif sigmma == 1.0:
            self.kernel = self.kernel_1.to(x.device)
            self.kernel_tran = self.kernel_tran_1.to(x.device)
            self.n_padding = 2
        elif sigmma == 1.5:
            self.kernel = self.kernel_15.to(x.device)
            self.kernel_tran = self.kernel_tran_15.to(x.device)
            self.n_padding = 3
        elif sigmma == 2:
            self.kernel = self.kernel_2.to(x.device)
            self.kernel_tran = self.kernel_tran_2.to(x.device)
            self.n_padding = 5
            self.n_padding = 4        # Multiscale inference
        elif sigmma == 4:
            self.kernel = self.kernel_4.to(x.device)
            self.kernel_tran = self.kernel_tran_4.to(x.device)
            self.n_padding = 10
        elif sigmma == 5:
            self.kernel = self.kernel_5.to(x.device)
            self.kernel_tran = self.kernel_tran_5.to(x.device)
            self.n_padding = 12
        elif sigmma == 6:
            self.kernel = self.kernel_6.to(x.device)
            self.kernel_tran = self.kernel_tran_6.to(x.device)
            self.n_padding = 15

        z_temp = []
        z = torch.zeros_like(x)
        
        for i in range(x.size()[1]):
            wd = self.n_padding
            x_pad = F.pad(x[:,i].unsqueeze(0),(wd,wd,wd,wd),'reflect')
            x_temp = F.conv2d(x_pad,self.kernel)
            
            y_temp = F.conv2d(x_temp,self.kernel_tran)
            z_temp.append(y_temp)
            
        z = torch.cat(z_temp, dim=1)

        return z   

class Diff(nn.Module):
    def __init__(self):
        super(Diff,self).__init__()

    def forward(self, x, direction):
        
        if direction == 'h':
            grad_temp = x[:,:,1:,:] - x[:,:,:-1,:]
            grad = F.pad(grad_temp,(0,0,0,1))

        elif direction == 'w':
            grad_temp = x[:,:,:,1:] - x[:,:,:,:-1]
            grad = F.pad(grad_temp,(0,1,0,0))
        return grad



class RTVW(nn.Module):
    def __init__(self):
        super(RTVW,self).__init__()

    def forward(self, x, eps):
        diff = Diff()
        grad_h = diff(x, 'h')
        grad_w = diff(x, 'w')
        
        w_temp = (grad_h.pow(2) + grad_w.pow(2)).clamp(min=(eps*0.1)**2)   
        
        w_temp = torch.pow(w_temp,0.5).sum(dim=1,keepdim=True)/x.size(1)     
        weight = torch.pow(weight, -1)
        return weight

class RTVU(nn.Module):
    def __init__(self):
        super(RTVU,self).__init__()
    def forward(self, x, sigmma, ep):
        gauss = GaussianFilter()
        y = gauss(x, sigmma)
        dif = Diff()
        y_h_temp = dif(y, 'h')
        y_h_temp = torch.abs(y_h_temp).sum(dim=1,keepdim=True)/x.size(1)     
        y_h = y_h_temp.clamp(min=ep)
        y_h = torch.pow(y_h, -1)

        y_w_temp = dif(y, 'w')
        y_w_temp = torch.abs(y_w_temp).sum(dim=1,keepdim=True)/x.size(1)
        y_w = y_w_temp.clamp(min=ep)
        y_w = torch.pow(y_w, -1)
        return y_h, y_w
     

class RTVLoss(nn.Module):
    def __init__(self):
        super(RTVLoss,self).__init__()

    def forward(self, x, sigmma, ep, eps):
        
        u = RTVU()
        weight_u_h, weight_u_w = u(x, sigmma, ep)

        w = RTVW()
        weight_rtvw = w(x, eps)
        dif = Diff()
        x_h = dif(x,'h')
        x_w = dif(x,'w')
        
        weight_h = weight_u_h.mul(weight_rtvw)
        weight_w = weight_u_w.mul(weight_rtvw) 
    
        loss_temp = weight_h.mul(torch.pow(x_h,2)) + weight_w.mul(torch.pow(x_w,2))
        loss = loss_temp.mean()
        return loss

class RTVSumLoss(nn.Module):
    def __init__(self):
        super(RTVSumLoss,self).__init__()

    def forward(self, x, sigmma, ep, eps):
        u = RTVU()
        weight_u_h, weight_u_w = u(x, sigmma, ep)
        w = RTVW()
        weight_rtvw = w(x, eps)
        dif = Diff()
        x_h = dif(x,'h')
        x_w = dif(x,'w')

        
        weight_h = weight_u_h.mul(weight_rtvw)
        weight_h = weight_h/weight_h.norm()
        weight_w = weight_u_w.mul(weight_rtvw) 
        weight_w = weight_w/weight_w.norm()
        
        loss_temp = weight_h.mul(torch.pow(x_h,2)) + weight_w.mul(torch.pow(x_w,2))
     
        loss = loss_temp.sum()
        return loss


class WLSLoss(nn.Module):
    def __init__(self):
        super(WLSLoss,self).__init__()
    def forward(self,I,x,alpha=1.2):
        epsilon = 1e-5
        I_log = torch.log(torch.clamp(I,min=epsilon,max=1.0))

        dif = Diff()
        I_h = dif(I_log,'h')
        I_w = dif(I_log,'w')
       
        x_h = dif(x,'h')
        x_w = dif(x,'w')
        weight_h = (I_h.abs() ** alpha + epsilon)**(-1)
        weight_w = (I_w.abs() ** alpha + epsilon)**(-1)
        loss_temp = weight_h.mul(x_h**2) + weight_w.mul(x_w**2)
        return loss_temp.mean()


class StructureLoss(nn.Module):
    # SSIM maximize
    def __init__(self,window_size=11,metric='SSIM'):
        super(StructureLoss,self).__init__()
        self.diff = Diff()
        if metric=='SSIM':
            self.ssim_loss = pytorch_ssim.SSIM(window_size)
            
        elif metric=='MSSSIM':
            self.ssim_loss = pytorch_msssim.MSSSIM(window_size)
        

    def forward(self,R,mask,window_size=11):                    
        R_h = self.diff(R,'h')
        R_w = self.diff(R,'w')
        Rgrad = (R_h.abs() + R_w.abs())/2
        if(mask.size(1)==1):
            mask = mask.repeat(1,3,1,1)
        
        loss = -self.ssim_loss(Rgrad, mask)
        return (loss).mean()                                      