import torch
import torch.nn as nn
from .layers import *
import torch.nn.functional as F
import math
from torch.autograd import Variable


def hard_sigmoid(x):
    return torch.min(torch.max(x, torch.zeros_like(x)), torch.ones_like(x))

class down(nn.Module):
    def __init__(self, in_ch, out_ch, filt, skip, pool_L,pool_H, if_original=True):
        super(down, self).__init__()
        self.conv = nn.Sequential(
            conv(in_ch, out_ch, filt, 1),
            nn.BatchNorm2d(out_ch),
            nn.LeakyReLU(inplace=True),
            conv(out_ch, out_ch, filt, 1),
            nn.BatchNorm2d(out_ch),
            nn.LeakyReLU(inplace=True)
        )
        self.pool = WavePool(out_ch,pool_L,pool_H)
        self.skip = skip
        self.if_original = if_original
        
    def forward(self, x):
        x = self.conv(x)
        L,LH,HL,HH = self.pool(x)
        y = None
        H = []
        if self.skip != 0:
            if self.if_original:
                y = x
            H = [LH, HL, HH]
        return L, y, H

class Diff(nn.Module):
    def __init__(self):
        super(Diff,self).__init__()

    def forward(self, x, direction):
        # if x.size(1) != 1:
        #     x = x.sum(dim=1,keepdim=True) / x.size(1)
        if direction == 'h':
            grad_temp = x[:,:,1:,:] - x[:,:,:-1,:]
            grad = F.pad(grad_temp,(0,0,0,1))

        elif direction == 'w':
            grad_temp = x[:,:,:,1:] - x[:,:,:,:-1]
            grad = F.pad(grad_temp,(0,1,0,0))
        return grad

class up(nn.Module):
    def __init__(self, in_ch, out_ch, filt, skip, skip_inchannel, pool_L, pool_H, bilinear=False):
        super(up, self).__init__()

        #  would be a nice idea if the upsampling could be learned too,
        #  but my machine do not have enough memory to handle all those weights
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose2d(in_ch//2, in_ch//2, 2, stride=2)

        self.unpool = WaveUnpool(in_ch, skip, skip_inchannel,pool_L,pool_H)
        if skip == 0:
            self.cha = in_ch
        elif skip == 1:
            # self.cha = in_ch * 3
            self.cha = in_ch + skip_inchannel * 4
        else:
            self.cha = in_ch + skip
        
        self.conv = nn.Sequential(
            conv(self.cha, out_ch, filt),
            nn.BatchNorm2d(out_ch),
            nn.LeakyReLU(inplace=True),
            conv(out_ch, out_ch, 1),
            nn.BatchNorm2d(out_ch),
            nn.LeakyReLU(inplace=True)
        )
        self.out_ch = out_ch
        self.filt = filt
    
    def forward(self, x, x_H, x_y):

        x = self.unpool(x, x_H, x_y)

        x = self.conv(x)
        return x

class L0(nn.Module):
    def __init__(self,
        crop_image_size, loc_mean=0, loc_sdev=0.01, beta=2. / 3., gamma=-0.1, zeta=1.1, fix_temp=True):
        super(L0, self).__init__()
        self._size = [1,1,crop_image_size[2],crop_image_size[3]]
        self.loc = nn.Parameter(torch.zeros(self._size).normal_(loc_mean, loc_sdev))
        self.temp = beta if fix_temp else nn.Parameter(torch.zeros(self._size).fill_(beta))
        self.register_buffer("uniform", torch.zeros(self._size))
        self.gamma = gamma
        self.zeta = zeta
        self.gamma_zeta_ratio = math.log(-gamma / zeta)

    def _get_mask(self):
        self.uniform.uniform_()
        u = Variable(self.uniform)
        u = u.clamp(min=1e-6)
        s = F.sigmoid((torch.log(u) - torch.log(1 - u) + self.loc) / self.temp)
        s = s * (self.zeta - self.gamma) + self.gamma

        penalty = F.sigmoid(self.loc - self.temp * self.gamma_zeta_ratio)
        return hard_sigmoid(s), penalty

    def get_grad(self,x):
        grad_temp_h = x[:,:,1:,:] - x[:,:,:-1,:]
        grad_h = F.pad(grad_temp_h,(0,0,0,1))

        grad_temp_w = x[:,:,:,1:] - x[:,:,:,:-1]
        grad_w = F.pad(grad_temp_w,(0,1,0,0))
        grad = (grad_h.abs() + grad_w.abs())/2
        grad = grad.sum(dim=1,keepdim=True)/3
        return grad.detach()

    def L0_mask_weight(self, image, alpha):
        dif = Diff()
        weight_h = dif(image,'h').abs().clamp(min=0.01).pow(-1)
        weight_w = dif(image,'w').abs().clamp(min=0.01).pow(-1)
        return (weight_h + weight_w)/2

    def forward(self, image, weight, ifdetach):
        self.temp = weight.detach()
        l0_mask,penalty = self._get_mask()
        if (ifdetach):
            weight = weight.detach()
        penalty = penalty * weight
        return l0_mask, penalty.mean(), 1-weight


class Rnet_wavelet(nn.Module):
    '''
    tongyong
    '''
    def __init__(self,
        crop_image_size,
        
        num_input_channels=2, num_output_channels=3, num_channels_down=[16, 32, 64, 128, 128],
        num_channels_up=[16, 32, 64, 128, 128],
        num_channels_skip=[4, 4, 4, 4, 4], pool_L=[2,2,2,2,2], pool_H = [2,2,2,2,2], filter_size_down=3,
        filter_size_up=3, filter_skip_size=1, need_sigmoid=True, need_bias=True,pad='reflection',
        upsample_mode='nearest', downsample_mode='stride', act_fun='LeakyReLU', need1x1_up=True,
        loc_mean=0, loc_sdev=0.01, beta=2. / 3., gamma=-0.1, zeta=1.1, fix_temp=True,
        ):
        super(Rnet_wavelet, self).__init__()
        assert len(num_channels_down) == len(num_channels_up) == len(num_channels_skip)

        n_scales = len(num_channels_down)
        if not (isinstance(upsample_mode, list) or isinstance(upsample_mode, tuple)):
            upsample_mode = [upsample_mode] * n_scales

        if not (isinstance(downsample_mode, list) or isinstance(downsample_mode, tuple)):
            downsample_mode = [downsample_mode] * n_scales

        if not (isinstance(filter_size_down, list) or isinstance(filter_size_down, tuple)):
            filter_size_down = [filter_size_down] * n_scales

        if not (isinstance(filter_size_up, list) or isinstance(filter_size_up, tuple)):
            filter_size_up = [filter_size_up] * n_scales
        
        self.down1 = down(num_input_channels, num_channels_down[0], filter_size_down[0], num_channels_skip[0],pool_L[0],pool_H[0])
        self.down2 = down(num_channels_down[0], num_channels_down[1],filter_size_down[1], num_channels_skip[1],pool_L[1],pool_H[1])
        self.down3 = down(num_channels_down[1], num_channels_down[2],filter_size_down[2], num_channels_skip[2],pool_L[2],pool_H[2])
        self.down4 = down(num_channels_down[2], num_channels_down[3],filter_size_down[3], num_channels_skip[3],pool_L[3],pool_H[3])
        self.down5 = down(num_channels_down[3], num_channels_down[4],filter_size_down[4], num_channels_skip[4],pool_L[4],pool_H[4])

        self.up1 = up(num_channels_up[1], num_channels_up[0], filter_size_up[0], num_channels_skip[0], num_channels_down[0],pool_L[0],pool_H[0])
        self.up2 = up(num_channels_up[2], num_channels_up[1], filter_size_up[1], num_channels_skip[1], num_channels_down[1],pool_L[1],pool_H[1])
        self.up3 = up(num_channels_up[3], num_channels_up[2], filter_size_up[2], num_channels_skip[2], num_channels_down[2],pool_L[2],pool_H[2])
        self.up4 = up(num_channels_up[4], num_channels_up[3], filter_size_up[3], num_channels_skip[3], num_channels_down[3],pool_L[3],pool_H[3])
        self.up5 = up(num_channels_down[4], num_channels_up[4], filter_size_up[4], num_channels_skip[4], num_channels_down[4],pool_L[4],pool_H[4])
        self.outc = nn.Sequential(nn.Conv2d(num_channels_up[0],num_output_channels,1), nn.Sigmoid())
        
    def forward(self, x):
        # 如果 skip=0，self.xn_y = None； self.xn_H = []，空list
        self.x1_L, self.x1_y, self.x1_H = self.down1(x) 
        self.x2_L, self.x2_y, self.x2_H = self.down2(self.x1_L)
        self.x3_L, self.x3_y, self.x3_H = self.down3(self.x2_L)
        self.x4_L, self.x4_y, self.x4_H = self.down4(self.x3_L)
        self.x5_L, self.x5_y, self.x5_H = self.down5(self.x4_L)

        self.y5 = self.up5(self.x5_L, self.x5_H, self.x5_y)
        self.y4 = self.up4(self.y5, self.x4_H, self.x4_y)
        self.y3 = self.up3(self.y4, self.x3_H, self.x3_y)
        self.y2 = self.up2(self.y3, self.x2_H, self.x2_y)
        self.y1 = self.up1(self.y2, self.x1_H, self.x1_y)
        # sigmoid
        self.y = self.outc(self.y1)
        return self.y




class Snet_wavelet(nn.Module):
    def __init__(self, n_channels=2, n_classes=3):
        super(Snet_wavelet, self).__init__()
        self.inc = inconv(n_channels, 64)
        self.down1 = down(64, 128)
        self.down2 = down(128, 256)
        self.down3 = down(256, 512)
        self.down4 = down(512, 512)
        self.up1 = up(1024, 256)
        self.up2 = up(512, 128)
        self.up3 = up(256, 64)
        self.up4 = up(128, 64)
        self.outc = outconv(64, n_classes)

    def forward(self, x):
        self.x1 = self.inc(x)
        self.x2 = self.down1(self.x1)
        self.x3 = self.down2(self.x2)
        self.x4 = self.down3(self.x3)
        self.x5 = self.down4(self.x4)
        self.x6 = self.up1(self.x5, self.x4)
        self.x7 = self.up2(self.x6, self.x3)
        self.x8 = self.up3(self.x7, self.x2)
        self.x9 = self.up4(self.x8, self.x1)
        self.y = self.outc(self.x9)
        return self.y