
import glob
import os
import torch
import torchvision
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image,ImageFilter
#import skvideo.io

matplotlib.use('agg')


def crop_image(img, d=32):
    """
    Make dimensions divisible by d

    :param pil img:
    :param d:
    :return:
    """

    new_size = (img.size[0] - img.size[0] % d,
                img.size[1] - img.size[1] % d)

    bbox = [
        int((img.size[0] - new_size[0]) / 2),
        int((img.size[1] - new_size[1]) / 2),
        int((img.size[0] + new_size[0]) / 2),
        int((img.size[1] + new_size[1]) / 2),
    ]

    img_cropped = img.crop(bbox)
    return img_cropped

def resize_image(img, d=32):
    """
    Make dimensions divisible by d

    :param pil img:
    :param d:
    :return:
    """

    if (img.size[0]//d + 1)*d - img.size[0] > img.size[0]%d:
        img_size0 = img.size[0] - img.size[0] % d
    else:
        img_size0 = (img.size[0]//d + 1)*d
        
    if (img.size[1]//d + 1)*d - img.size[1] > img.size[1]%d:
        img_size1 = img.size[1] - img.size[1] % d
    else:
        img_size1 = (img.size[1]//d + 1)*d

    img_resized = img.resize((img_size0, img_size1), Image.ANTIALIAS)
    return img_resized, img.size[0],img.size[1]


def crop_np_image(img_np, d=32):
    return torch_to_np(crop_torch_image(np_to_torch(img_np), d))


def crop_torch_image(img, d=32):
    """
    Make dimensions divisible by d
    image is [1, 3, W, H] or [3, W, H]
    :param pil img:
    :param d:
    :return:
    """
    new_size = (img.shape[-2] - img.shape[-2] % d,
                img.shape[-1] - img.shape[-1] % d)
    pad = ((img.shape[-2] - new_size[-2]) // 2, (img.shape[-1] - new_size[-1]) // 2)

    if len(img.shape) == 4:
        return img[:, :, pad[-2]: pad[-2] + new_size[-2], pad[-1]: pad[-1] + new_size[-1]]
    assert len(img.shape) == 3
    return img[:, pad[-2]: pad[-2] + new_size[-2], pad[-1]: pad[-1] + new_size[-1]]


def get_params(opt_over, net, net_input, downsampler=None):
    """
    Returns parameters that we want to optimize over.
    :param opt_over: comma separated list, e.g. "net,input" or "net"
    :param net: network
    :param net_input: torch.Tensor that stores input `z`
    :param downsampler:
    :return:
    """

    opt_over_list = opt_over.split(',')
    params = []

    for opt in opt_over_list:

        if opt == 'net':
            params += [x for x in net.parameters()]
        elif opt == 'down':
            assert downsampler is not None
            params = [x for x in downsampler.parameters()]
        elif opt == 'input':
            net_input.requires_grad = True
            params += [net_input]
        else:
            assert False, 'what is it?'

    return params


def get_image_grid(images_np, nrow=8):
    """
    Creates a grid from a list of images by concatenating them.
    :param images_np:
    :param nrow:
    :return:
    """
    images_torch = [torch.from_numpy(x).type(torch.FloatTensor) for x in images_np]
    torch_grid = torchvision.utils.make_grid(images_torch, nrow)

    return torch_grid.numpy()


def plot_image_grid(name, images_np, interpolation='lanczos', output_path="output/"):
    """
    Draws images in a grid

    Args:
        images_np: list of images, each image is np.array of size 3xHxW or 1xHxW
        nrow: how many images will be in one row
        interpolation: interpolation used in plt.imshow
    """
    assert len(images_np) == 2 
    n_channels = max(x.shape[0] for x in images_np)
    assert (n_channels == 3) or (n_channels == 1), "images should have 1 or 3 channels"

    images_np = [x if (x.shape[0] == n_channels) else np.concatenate([x, x, x], axis=0) for x in images_np]

    grid = get_image_grid(images_np, 2)

    if images_np[0].shape[0] == 1:
        plt.imshow(grid[0], cmap='gray', interpolation=interpolation)
    else:
        plt.imshow(grid.transpose(1, 2, 0), interpolation=interpolation)

    plt.savefig(output_path + "{}.png".format(name))


def save_image(name, image_np, width, height,output_path="output"):
    path = os.path.join(output_path, name+'.png')
    p = np_to_pil(image_np)
    p = p.resize((width,height),Image.ANTIALIAS)
    p.save(path)

def video_to_images(file_name, name):
    video = prepare_video(file_name)
    for i, f in enumerate(video):
        save_image(name + "_{0:03d}".format(i), f)

def images_to_video(images_dir ,name, gray=True):
    num = len(glob.glob(images_dir +"/*.png"))
    c = []
    for i in range(num):
        if gray:
            img = prepare_gray_image(images_dir + "/"+  name +"_{}.png".format(i))
        else:
            img = prepare_image(images_dir + "/"+name+"_{}.png".format(i))
        print(img.shape)
        c.append(img)
    save_video(name, np.array(c))

def save_heatmap(name, image_np):
    cmap = plt.get_cmap('jet')

    rgba_img = cmap(image_np)
    rgb_img = np.delete(rgba_img, 3, 2)
    save_image(name, rgb_img.transpose(2, 0, 1))


def save_graph(name, graph_list, output_path="output"):
    path = os.path.join(output_path, name+'.png')
    plt.clf()
    plt.plot(graph_list)
    plt.savefig(path)


def create_augmentations(np_image):
    """
    convention: original, left, upside-down, right, rot1, rot2, rot3
    :param np_image:
    :return:
    """
    aug = [np_image.copy(), np.rot90(np_image, 1, (1, 2)).copy(),
           np.rot90(np_image, 2, (1, 2)).copy(), np.rot90(np_image, 3, (1, 2)).copy()]
    flipped = np_image[:,::-1, :].copy()
    aug += [flipped.copy(), np.rot90(flipped, 1, (1, 2)).copy(), np.rot90(flipped, 2, (1, 2)).copy(), np.rot90(flipped, 3, (1, 2)).copy()]
    return aug


def create_video_augmentations(np_video):
    """
        convention: original, left, upside-down, right, rot1, rot2, rot3
        :param np_video:
        :return:
        """
    aug = [np_video.copy(), np.rot90(np_video, 1, (2, 3)).copy(),
           np.rot90(np_video, 2, (2, 3)).copy(), np.rot90(np_video, 3, (2, 3)).copy()]
    flipped = np_video[:, :, ::-1, :].copy()
    aug += [flipped.copy(), np.rot90(flipped, 1, (2, 3)).copy(), np.rot90(flipped, 2, (2, 3)).copy(),
            np.rot90(flipped, 3, (2, 3)).copy()]
    return aug


def save_graphs(name, graph_dict, output_path="output/"):
    """

    :param name:
    :param dict graph_dict: a dict from the name of the list to the list itself.
    :return:
    """
    plt.clf()
    fig, ax = plt.subplots()
    for k, v in graph_dict.items():
        ax.plot(v, label=k)
        # ax.semilogy(v, label=k)
    ax.set_xlabel('iterations')
    # ax.set_ylabel(name)
    ax.set_ylabel('MSE-loss')
    # ax.set_ylabel('PSNR')
    plt.legend()
    plt.savefig(output_path + name + ".png")


def load(path):
    """Load PIL image."""
    img = Image.open(path)#.convert('RGB')
    return img


def get_image(path, imsize=-1):
    """Load an image and resize to a cpecific size.

    Args:
        path: path to image
        imsize: tuple or scalar with dimensions; -1 for `no resize`
    """
    img = load(path)

    if isinstance(imsize, int):
        imsize = (imsize, imsize)

    if imsize[0] != -1 and img.size != imsize:
        if imsize[0] > img.size[0]:
            img = img.resize(imsize, Image.BICUBIC)
        else:
            img = img.resize(imsize, Image.ANTIALIAS)

    img_np = pil_to_np(img)

    return img, img_np

def rgb_to_irg(rgb):
    """ converts rgb to (mean of channels, red chromaticity, green chromaticity) """
    irg = np.zeros_like(rgb)
    s = np.sum(rgb, axis=-1) + 1e-6

    irg[..., 2] = s / 3.0
    irg[..., 0] = rgb[..., 0] / s
    irg[..., 1] = rgb[..., 1] / s
    return irg

def srgb_to_rgb(srgb):
    ret = np.zeros_like(srgb)
    idx0 = srgb <= 0.04045
    idx1 = srgb > 0.04045
    ret[idx0] = srgb[idx0] / 12.92
    ret[idx1] = np.power((srgb[idx1] + 0.055) / 1.055, 2.4)
    return ret


def rgb_to_chromaticity(rgb):
    """ converts rgb to chromaticity """
    irg = np.zeros_like(rgb)
    s = np.sum(rgb, axis=-1) + 1e-6

    irg[..., 0] = rgb[..., 0] / s
    irg[..., 1] = rgb[..., 1] / s
    irg[..., 2] = rgb[..., 2] / s
    return irg    
    
def construst_R_weights(N_feature,sigma_chro,sigma_I):
    
    dim = N_feature.shape[0]
    center_feature = np.repeat( np.expand_dims(N_feature[(dim+1)//2, :, :,:], axis =0), dim , axis = 0) 
    feature_diff = center_feature - N_feature

    r_w = np.exp( - np.sum( feature_diff[:,:,:,0:2]**2  , 3) / (sigma_chro**2)) \
                * np.exp(- (feature_diff[:,:,:,2]**2) /(sigma_I**2) )
    return r_w

def construst_sub_matrix(C,win_r):
    h = C.shape[0]
    w = C.shape[1]
    half_window = win_r                               
    x = np.arange(-win_r, win_r+1)                     
    y = np.arange(-win_r, win_r+1)
    X,Y = np.meshgrid(x,y)

    sub_C = np.zeros( ((2*half_window+1)**2 ,h-2*win_r,w-2*win_r, 3))        
    ct_idx = 0
    for k in range(0, half_window*2+1):
        for l in range(0, half_window*2+1):
            sub_C[ct_idx,:,:,:] = C[half_window + Y[k,l] : h - half_window + Y[k,l], \
            half_window + X[k,l]: w - half_window + X[k,l] , :] 
            ct_idx += 1
    return sub_C


def prepare_image(file_name,d=32):
    """
    loads makes it divisible
    :param file_name:
    :return: the numpy representation of the image
    """
    img_pil,width,height = resize_image(get_image(file_name, -1)[0], d)
    return pil_to_np(img_pil),width,height


def prepare_image_and_weight(file_name, srgb=True, eps=1e-5, num_scale=4, d=32,win_r=1,sigma_chro=0.025,sigma_I=0.1):
    """
    """
    img_pil,width,height = resize_image(get_image(file_name, -1)[0], d)
    # print(img_pil.size)
    np_srgb = pil_to_np(img_pil)                    
    # print(np_srgb.shape)                           
    if srgb:
        img = np_srgb
    else:
        img = srgb_to_rgb(np_srgb)  

    img[img < eps] = eps                             
    rgb_img = img.transpose(1,2,0)                   
    R_weight = []
    for i in range(0, num_scale):
        feature_3d = rgb_to_irg(rgb_img)
        
        sub_matrix = construst_sub_matrix(feature_3d,win_r)
        
        r_w = construst_R_weights(sub_matrix,sigma_chro,sigma_I)
        
        R_weight.append(torch.from_numpy(r_w).float().unsqueeze(0))        
        
        rgb_img = rgb_img[::2,::2,:]
   
    return img, R_weight, width, height

def multiscale_prepare_image(file_name,minsize_list,net_depth):
    """
    loads makes it divisible
    :param file_name:
    :return: the numpy representation of the image
    """
    img = get_image(file_name, -1)[0]              
    
    min_size = min(img.size[0],img.size[1])
    temp_list = [abs(number-min_size) for number in minsize_list]
    list_clip = temp_list.index(min(temp_list))
    new_list = minsize_list[0:list_clip+1]
    d = net_depth[0:list_clip+1]

    
    max_size = max(img.size[0],img.size[1])        # size[0] 是 w，size[1] 是 H
    width = []
    height = []
    img_resized = []
    for i in range(len(new_list)):
        temp_min = minsize_list[i]
        temp_max = round(max_size / min_size * temp_min)

        if (temp_max//(2**d[i]) + 1)*(2**d[i]) - temp_max > temp_max%(2**d[i]):
            temp_max = temp_max - temp_max%(2**d[i])
        else:
            temp_max = (temp_max//(2**d[i]) + 1)*(2**d[i])
        if img.size[0] > img.size[1]:
            temp_h = temp_min
            temp_w = temp_max
        else:
            temp_h = temp_max
            temp_w = temp_min

        temp_img = img.resize((temp_w, temp_h), Image.ANTIALIAS)        
        np_img = pil_to_np(temp_img)
        width.append(temp_w)
        height.append(temp_h)
        img_resized.append(np_img)
    height.append(img.size[1])
    width.append(img.size[0])
    return img_resized,width,height


def multiscale_gauss_prepare_image(file_name,minsize_list,net_depth,radius):
    """
    loads makes it divisible
    :param file_name:
    :return: the numpy representation of the image
    """
    # img_pil,width,height = resize_image(get_image(file_name, -1)[0], d)
    img = get_image(file_name, -1)[0]      
    
    min_size = min(img.size[0],img.size[1])
    temp_list = [abs(number-min_size) for number in minsize_list]
    list_clip = temp_list.index(min(temp_list))
    new_list = minsize_list[0:list_clip+1]
    d = net_depth[0:list_clip+1]

    max_size = max(img.size[0],img.size[1])        # size[0] 是 w，size[1] 是 H
    width = []
    height = []
    img_resized = []
    temp_img = img
    for j in range(len(new_list)):
        i = - (j+1)
        temp_min = minsize_list[i]
        temp_max = round(max_size / min_size * temp_min)

        if (temp_max//(2**d[i]) + 1)*(2**d[i]) - temp_max > temp_max%(2**d[i]):
            temp_max = temp_max - temp_max%(2**d[i])
        else:
            temp_max = (temp_max//(2**d[i]) + 1)*(2**d[i])
        if img.size[0] > img.size[1]:
            temp_h = temp_min
            temp_w = temp_max
        else:
            temp_h = temp_max
            temp_w = temp_min
        if j != 0:
            temp_img = temp_img.filter(ImageFilter.GaussianBlur(radius)) 
        temp_img = temp_img.resize((temp_w, temp_h), Image.ANTIALIAS)
        np_img = pil_to_np(temp_img)
        width.insert(0,temp_w)
        height.insert(0,temp_h)
        img_resized.insert(0,np_img)
    height.append(img.size[1])
    width.append(img.size[0])
    return img_resized,width,height


def construst_R_weights_myself(N_feature,sigma_chro,sigma_I):
    dim = N_feature.shape[0]
    center_feature = np.repeat( np.expand_dims(N_feature[(dim+1)//2, :, :,:], axis =0), dim , axis = 0) 
    # center_feature = np.repeat( np.expand_dims(N_feature[4, :, :,:], axis =0), 9 , axis = 0)     
    feature_diff = center_feature - N_feature

    r_w = np.exp( - np.sum( feature_diff[:,:,:,0:2]**2  , 3) / (sigma_chro**2)) \
                * np.exp(- (feature_diff[:,:,:,2]**2) /(sigma_I**2) )
    
    return r_w

def construst_sub_matrix_myself(C,win_r):
    h = C.shape[0]
    w = C.shape[1]
    half_window = win_r                                
    x = np.arange(-win_r, win_r+1)                    
    y = np.arange(-win_r, win_r+1)
    X,Y = np.meshgrid(x,y)

    sub_C = np.zeros( ((2*half_window+1)**2 ,h,w, 3))         
    ct_idx = 0
    C = np.pad(C,((win_r,win_r),(win_r,win_r),(0,0)),'reflect')      
    for k in range(0, half_window*2+1):
        for l in range(0, half_window*2+1):
            sub_C[ct_idx,:,:,:] = C[half_window+Y[k,l] : (h + half_window+ Y[k,l]), half_window + X[k,l]: (w + half_window+ X[k,l]) , :] 
            ct_idx += 1
    return sub_C


def prepare_image_myself(file_name,d=32):
    """
    loads makes it divisible
    :param file_name:
    :return: the numpy representation of the image
    """
    img_pil,width,height = resize_image(get_image(file_name, -1)[0], d)
    return pil_to_np(img_pil),width,height


def prepare_image_and_weight_v2(file_name, srgb=True, eps=1e-5, num_scale=4, d=32,win_r=1,sigma_chro=0.025,sigma_I=0.1):
    """
    """
    img_pil,width,height = resize_image(get_image(file_name, -1)[0], d)
   
    np_srgb = pil_to_np(img_pil)                   
                   
    if srgb:
        img = np_srgb
    else:
        img = srgb_to_rgb(np_srgb)  

    img[img < eps] = eps                             
    rgb_img = img.transpose(1,2,0)                   
    R_weight = []
    for i in range(0, num_scale):
        feature_3d = rgb_to_irg(rgb_img)
       
        sub_matrix = construst_sub_matrix_myself(feature_3d, win_r)
       
        r_w = construst_R_weights_myself(sub_matrix, sigma_chro, sigma_I)
       
        R_weight.append(torch.from_numpy(r_w).float().unsqueeze(0).cuda())
        
        rgb_img = rgb_img[::2,::2,:]
    return img, R_weight, width, height




# def prepare_video(file_name, folder="output/"):
    # data = skvideo.io.vread(folder + file_name)
    # return crop_torch_image(data.transpose(0, 3, 1, 2).astype(np.float32) / 255.)[:35]


# def save_video(name, video_np, output_path="output/"):
    # outputdata = video_np * 255
    # outputdata = outputdata.astype(np.uint8)
    # skvideo.io.vwrite(output_path + "{}.mp4".format(name), outputdata.transpose(0, 2, 3, 1))


def prepare_gray_image(file_name):
    img = prepare_image(file_name)
    return np.array([np.mean(img, axis=0)])


def pil_to_np(img_PIL, with_transpose=True):
    """
    Converts image in PIL format to np.array.

    From W x H x C [0...255] to C x W x H [0..1]
 
    """
    ar = np.array(img_PIL)
    if len(ar.shape) == 3 and ar.shape[-1] == 4:
        ar = ar[:, :, :3]
        # this is alpha channel
    if with_transpose:
        if len(ar.shape) == 3:
            ar = ar.transpose(2, 0, 1)
        else:
            ar = ar[None, ...]

    return ar.astype(np.float32) / 255.


def median(img_np_list):
    """
    assumes C x W x H [0..1]
    :param img_np_list:
    :return:
    """
    assert len(img_np_list) > 0
    l = len(img_np_list)
    shape = img_np_list[0].shape
    result = np.zeros(shape)
    for c in range(shape[0]):
        for w in range(shape[1]):
            for h in range(shape[2]):
                result[c, w, h] = sorted(i[c, w, h] for i in img_np_list)[l//2]
    return result


def average(img_np_list):
    """
    assumes C x W x H [0..1]
    :param img_np_list:
    :return:
    """
    assert len(img_np_list) > 0
    l = len(img_np_list)
    shape = img_np_list[0].shape
    result = np.zeros(shape)
    for i in img_np_list:
        result += i
    return result / l


def np_to_pil(img_np):
    """
    Converts image in np.array format to PIL image.

    From C x W x H [0..1] to  W x H x C [0...255]
    :param img_np:
    :return:
    """
    ar = np.clip(img_np * 255, 0, 255).astype(np.uint8)

    if img_np.shape[0] == 1:
        ar = ar[0]
    else:
        assert img_np.shape[0] == 3, img_np.shape
        ar = ar.transpose(1, 2, 0)                    

    return Image.fromarray(ar)


def np_to_torch(img_np):
    """
    Converts image in numpy.array to torch.Tensor.

    From C x W x H [0..1] to  C x W x H [0..1]

    :param img_np:
    :return:
    """
    return torch.from_numpy(img_np)[None, :]


def torch_to_np(img_var):
    """
    Converts an image in torch.Tensor format to np.array.

    From 1 x C x W x H [0..1] to  C x W x H [0..1]
    :param img_var:
    :return:
    """
    return img_var.detach().cpu().numpy()[0]
